/* pci-apm.c: Linux APM support code for pci-scan. */
/*
	Written 2000-2001 by Donald Becker.

	This software may be used and distributed according to the terms
	of the GNU General Public License (GPL), incorporated herein by
	reference.  This is not a documented interface.  Drivers
	incorporating or interacting with these functions are derivative
	works and thus are covered the GPL.  They must include an explicit
	GPL notice.

	This code provides an interface to push APM events to hot-swap drivers
	using the pci-scan code.
	The functions supported are suspend, power-off, and resume.

	The author may be reached as becker@scyld.com, or
	Donald Becker
	Scyld Computing Corporation
	410 Severn Ave., Suite 210
	Annapolis MD 21403

	Support and updates available at
	http://www.scyld.com/network/drivers.html

	Other contributers: (none yet)
*/

static const char version1[] =
"pci-apm.c:v1.01 1/31/2000  Donald Becker <becker@scyld.com>\n";
static const char version2[] =
" http://www.scyld.com/linux/drivers.html\n";

/* Module options. */
static int debug = 1;			/* 1 normal messages, 0 quiet .. 7 verbose. */

#ifndef __KERNEL__
#define __KERNEL__
#endif
#include <linux/config.h>
#if defined(CONFIG_SMP) && ! defined(__SMP__)
#define __SMP__
#endif
#if defined(MODULE) && defined(CONFIG_MODVERSIONS) && ! defined(MODVERSIONS)
#define MODVERSIONS
#endif

#include <linux/version.h>
#include <linux/module.h>
#if LINUX_VERSION_CODE < 0x20300  &&  defined(MODVERSIONS)
#include <linux/modversions.h>
#endif
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/pci.h>

/* These might be awkward to locate. */
#include <linux/pci-scan.h>
#include <linux/kern_compat.h>

MODULE_AUTHOR("Donald Becker <becker@scyld.com>");
MODULE_DESCRIPTION("APM support for hot-swap PCI");
MODULE_PARM(debug, "i");


#ifdef CONFIG_APM
#include <linux/apm_bios.h>

struct drv_shim {
	struct drv_shim *next;
	struct drv_id_info *did;
} static *root_drv_id = 0;

static int handle_apm_event(apm_event_t event)
{
	static int down = 0;
	struct drv_shim *dp;

	if (debug > 1)
		printk(KERN_DEBUG "pci-apm: Handling APM event %d.\n", event);
	switch (event) {
	case APM_SYS_SUSPEND:
	case APM_USER_SUSPEND:
		if (down) {
			printk(KERN_DEBUG "pci-apm: Received extra suspend event\n");
			break;
		}
		down = 1;
		for (dp = root_drv_id; dp; dp = dp->next)
			if (dp->did->pwr_event)
				dp->did->pwr_event(dp->did, DRV_SUSPEND);
		break;
	case APM_NORMAL_RESUME:
	case APM_CRITICAL_RESUME:
		if (!down) {
			printk(KERN_DEBUG "pci-apm: Received bogus resume event\n");
			break;
		}
		for (dp = root_drv_id; dp; dp = dp->next)
			if (dp->did->pwr_event)
				dp->did->pwr_event(dp->did, DRV_RESUME);
		down = 0;
		break;
	}
	return 0;
}

/* Add/remove a driver ID structure to our private list of known drivers. */
int do_cb_register(struct drv_id_info *did)
{
	struct drv_shim *dshim = kmalloc(sizeof(*dshim), GFP_KERNEL);
	if (dshim == 0)
		return -ENOMEM;
	if (debug > 1)
		printk(KERN_INFO "Registering APM driver support for '%s'.\n",
			   did->name);
	MOD_INC_USE_COUNT;
	dshim->did = did;
	dshim->next = root_drv_id;
	root_drv_id = dshim;
	return 0;
}

void do_cb_unregister(struct drv_id_info *did)
{
	struct drv_shim **dp;
	for (dp = &root_drv_id; *dp; dp = &(*dp)->next)
		if ((*dp)->did == did) {
			struct drv_shim *dshim = *dp;
			*dp = dshim->next;
			kfree(dshim);
			MOD_DEC_USE_COUNT;
			return;
		}
}

extern int (*register_cb_hook)(struct drv_id_info *did);
extern void (*unregister_cb_hook)(struct drv_id_info *did);

int (*old_cb_hook)(struct drv_id_info *did);
void (*old_un_cb_hook)(struct drv_id_info *did);

int init_module(void)
{
	if (debug)
		printk(KERN_INFO "%s" KERN_INFO "%s", version1, version2);
	old_cb_hook = register_cb_hook;
	old_un_cb_hook = unregister_cb_hook;
	register_cb_hook = do_cb_register;
	unregister_cb_hook = do_cb_unregister;
#if defined(CONFIG_APM)
	apm_register_callback(&handle_apm_event);
#endif
	return 0;
}
void cleanup_module(void)
{
#if defined(CONFIG_APM)
	apm_unregister_callback(&handle_apm_event);
#endif
	register_cb_hook = 	old_cb_hook;
	unregister_cb_hook = unregister_cb_hook;
	return;
}

#endif

/*
 * Local variables:
 *  compile-command: "gcc -DMODULE -Wall -Wstrict-prototypes -O6 -c pci-apm.c"
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 4
 * End:
 */
