/*
 *  linux/fs/open.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 */

#include <linux/mm.h>
#include <linux/utime.h>
#include <linux/file.h>
#include <linux/smp_lock.h>
#include <linux/quotaops.h>

#include <asm/uaccess.h>

/* RSBAC */
#ifdef CONFIG_RSBAC
#include <rsbac/adf.h>
#endif

asmlinkage long sys_statfs(const char * path, struct statfs * buf)
{
	struct dentry * dentry;
	int error;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
        enum  rsbac_target_t          rsbac_target;
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

	lock_kernel();
	dentry = namei(path);
	error = PTR_ERR(dentry);
	if (!IS_ERR(dentry)) {
		struct inode * inode = dentry->d_inode;
		struct super_block * sb = inode->i_sb;

                /* RSBAC */
                #ifdef CONFIG_RSBAC
#ifdef CONFIG_RSBAC_DEBUG
                if (rsbac_debug_aef) printk(KERN_DEBUG "sys_statfs(): calling ADF\n");
#endif
                rsbac_target = T_FILE;
                if (S_ISDIR(dentry->d_inode->i_mode))
                  rsbac_target = T_DIR;
                else if (S_ISFIFO(dentry->d_inode->i_mode))
                  rsbac_target = T_FIFO;
                else if (S_ISLNK(dentry->d_inode->i_mode))
                  rsbac_target = T_SYMLINK;
                rsbac_target_id.file.device = inode->i_dev;
                rsbac_target_id.file.inode  = inode->i_ino;
                rsbac_target_id.file.dentry_p = dentry;
                rsbac_attribute_value.dummy = 0;
                if (!rsbac_adf_request(R_GET_STATUS_DATA,
                                       current->pid,
                                       rsbac_target,
                                       rsbac_target_id,
                                       A_none,
                                       rsbac_attribute_value))
                  {
                    dput(dentry);
                    unlock_kernel();
                    return -EPERM;
                  }
                #endif
		error = -ENODEV;
		if (sb && sb->s_op && sb->s_op->statfs)
			error = sb->s_op->statfs(sb, buf, sizeof(struct statfs));

		dput(dentry);
	}
	unlock_kernel();
	return error;
}

asmlinkage long sys_fstatfs(unsigned int fd, struct statfs * buf)
{
	struct file * file;
	struct inode * inode;
	struct dentry * dentry;
	struct super_block * sb;
	int error;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
        enum  rsbac_target_t          rsbac_target;
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

	lock_kernel();
	error = -EBADF;
	file = fget(fd);
	if (!file)
		goto out;
	error = -ENOENT;
	if (!(dentry = file->f_dentry))
		goto out_putf;
	if (!(inode = dentry->d_inode))
		goto out_putf;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
#ifdef CONFIG_RSBAC_DEBUG
        if (rsbac_debug_aef) printk(KERN_DEBUG "sys_fstatfs(): calling ADF\n");
#endif
        rsbac_target = T_FILE;
        if (S_ISDIR(inode->i_mode))
          rsbac_target = T_DIR;
        else if (S_ISFIFO(inode->i_mode))
          rsbac_target = T_FIFO;
        else if (S_ISLNK(inode->i_mode))
          rsbac_target = T_SYMLINK;
        rsbac_target_id.file.device = inode->i_dev;
        rsbac_target_id.file.inode  = inode->i_ino;
        rsbac_target_id.file.dentry_p = dentry;
        rsbac_attribute_value.dummy = 0;
        if (!rsbac_adf_request(R_GET_STATUS_DATA,
                               current->pid,
                               rsbac_target,
                               rsbac_target_id,
                               A_none,
                               rsbac_attribute_value))
          {
            error = -EPERM;
            goto out_putf;
          }
        #endif

	error = -ENODEV;
	sb = inode->i_sb;
	if (sb && sb->s_op && sb->s_op->statfs)
		error = sb->s_op->statfs(sb, buf, sizeof(struct statfs));
out_putf:
	fput(file);
out:
	unlock_kernel();
	return error;
}

int do_truncate(struct dentry *dentry, loff_t length)
{
	struct inode *inode = dentry->d_inode;
	int error;
	struct iattr newattrs;
        /* RSBAC */
        #ifdef CONFIG_RSBAC_SECDEL
        off_t old_len = inode->i_size;
        #endif

	/* Not pretty: "inode->i_size" shouldn't really be signed. But it is. */
	error = -EINVAL;
	if (length < 0)
		goto out;

	fs_down(&inode->i_sem);
	newattrs.ia_size = length;
	newattrs.ia_valid = ATTR_SIZE | ATTR_CTIME;
	error = notify_change(dentry, &newattrs);
	if (!error) {
		/* truncate virtual mappings of this file */
		vmtruncate(inode, length);
                /* RSBAC: Overwrite truncated part, if asked by flag */
                #ifdef CONFIG_RSBAC_SECDEL
                rsbac_sec_trunc(dentry, length, old_len);
                newattrs.ia_size = length;
                newattrs.ia_valid = ATTR_SIZE | ATTR_CTIME;
                error = notify_change(dentry, &newattrs);
                if (!error)
                #endif
		if (inode->i_op && inode->i_op->truncate)
			inode->i_op->truncate(inode);
	}
	fs_up(&inode->i_sem);
out:
	return error;
}

static inline long do_sys_truncate(const char * path, loff_t length)
{
	struct dentry * dentry;
	struct inode * inode;
	int error;

	/* RSBAC */
        #ifdef CONFIG_RSBAC
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_target_id_t       rsbac_new_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

	lock_kernel();
	dentry = namei(path);

	error = PTR_ERR(dentry);
	if (IS_ERR(dentry))
		goto out;
	inode = dentry->d_inode;

	error = -EACCES;
	if (S_ISDIR(inode->i_mode))
		goto dput_and_out;

	error = permission(inode,MAY_WRITE);
	if (error)
		goto dput_and_out;

	error = -EROFS;
	if (IS_RDONLY(inode))
		goto dput_and_out;

	error = -EPERM;
	if (IS_IMMUTABLE(inode) || IS_APPEND(inode))
		goto dput_and_out;

	error = get_write_access(inode);
	if (error)
		goto dput_and_out;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
#ifdef CONFIG_RSBAC_DEBUG
        if (rsbac_debug_aef) printk(KERN_DEBUG "sys_truncate(): calling ADF\n");
#endif
        rsbac_target_id.file.device = inode->i_dev;
        rsbac_target_id.file.inode  = inode->i_ino;
        rsbac_target_id.file.dentry_p = dentry;
        rsbac_attribute_value.dummy = 0;
        if (!rsbac_adf_request(R_TRUNCATE,
                               current->pid,
                               T_FILE,
                               rsbac_target_id,
                               A_none,
                               rsbac_attribute_value))
          {
            put_write_access(inode);
            error = -EPERM;
            goto dput_and_out;
          }
        #endif

	error = locks_verify_area(FLOCK_VERIFY_WRITE, inode, NULL,
				  length < inode->i_size ? length : inode->i_size,
				  abs(inode->i_size - length));
	if (!error) {
		DQUOT_INIT(inode);
		error = do_truncate(dentry, length);

                #ifdef CONFIG_RSBAC
                if (!error)
                  {
#ifdef CONFIG_RSBAC_DEBUG
                    if (rsbac_debug_aef)
                      printk(KERN_DEBUG
                        "sys_truncate: notify ADF of FILE TRUNCATE\n");
#endif
                    rsbac_new_target_id.dummy = 0;
                    rsbac_attribute_value.dummy = 0;
                    if (rsbac_adf_set_attr(R_TRUNCATE,
                                           current->pid,
                                           T_FILE,
                                           rsbac_target_id,
                                           T_NONE,
                                           rsbac_new_target_id,
                                           A_none,
                                           rsbac_attribute_value))
                      {
                        printk(KERN_WARNING 
                               "sys_truncate(): rsbac_adf_set_attr() returned error\n");
                      }
                  }
                #endif
	}
	put_write_access(inode);
dput_and_out:
	dput(dentry);
out:
	unlock_kernel();
	return error;
}

asmlinkage long sys_truncate(const char * path, unsigned long length)
{
	return do_sys_truncate(path, length);
}

static inline long do_sys_ftruncate(unsigned int fd, loff_t length)
{
	struct inode * inode;
	struct dentry *dentry;
	struct file * file;
	int error;

	/* RSBAC */
        #ifdef CONFIG_RSBAC
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_target_id_t       rsbac_new_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

	lock_kernel();
	error = -EBADF;
	file = fget(fd);
	if (!file)
		goto out;
	error = -ENOENT;
	if (!(dentry = file->f_dentry))
		goto out_putf;
	if (!(inode = dentry->d_inode))
		goto out_putf;
	error = -EACCES;
	if (S_ISDIR(inode->i_mode) || !(file->f_mode & FMODE_WRITE))
		goto out_putf;
	error = -EPERM;
	if (IS_IMMUTABLE(inode) || IS_APPEND(inode))
		goto out_putf;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
#ifdef CONFIG_RSBAC_DEBUG
        if (rsbac_debug_aef) printk(KERN_DEBUG "sys_ftruncate(): calling ADF\n");
#endif
        rsbac_target_id.file.device = inode->i_dev;
        rsbac_target_id.file.inode  = inode->i_ino;
        rsbac_target_id.file.dentry_p = dentry;
        rsbac_attribute_value.dummy = 0;
        if (!rsbac_adf_request(R_TRUNCATE,
                               current->pid,
                               T_FILE,
                               rsbac_target_id,
                               A_none,
                               rsbac_attribute_value))
          {
            error = -EPERM;
            goto out_putf;
          }
        #endif

	error = locks_verify_area(FLOCK_VERIFY_WRITE, inode, file,
				  length<inode->i_size ? length : inode->i_size,
				  abs(inode->i_size - length));
	if (!error)
		error = do_truncate(dentry, length);

        #ifdef CONFIG_RSBAC
        if (!error)
          {
#ifdef CONFIG_RSBAC_DEBUG
            if (rsbac_debug_aef)
              printk(KERN_DEBUG
                     "sys_ftruncate: notify ADF of FILE TRUNCATE\n");
#endif
            rsbac_new_target_id.dummy = 0;
            rsbac_attribute_value.dummy = 0;
            if (rsbac_adf_set_attr(R_TRUNCATE,
                                   current->pid,
                                   T_FILE,
                                   rsbac_target_id,
                                   T_NONE,
                                   rsbac_new_target_id,
                                   A_none,
                                   rsbac_attribute_value))
              {
                printk(KERN_WARNING 
                       "sys_ftruncate(): rsbac_adf_set_attr() returned error\n");
              }
          }
        #endif
out_putf:
	fput(file);
out:
	unlock_kernel();
	return error;
}

asmlinkage long sys_ftruncate(unsigned int fd, unsigned long length)
{
	return do_sys_ftruncate(fd, length);
}

/* LFS versions of truncate are only needed on 32 bit machines */
#if BITS_PER_LONG == 32
asmlinkage long sys_truncate64(const char * path, loff_t length)
{
	return do_sys_truncate(path, length);
}

asmlinkage long sys_ftruncate64(unsigned int fd, loff_t length)
{
	return do_sys_ftruncate(fd, length);
}
#endif

#ifndef __alpha__

/*
 * sys_utime() can be implemented in user-level using sys_utimes().
 * Is this for backwards compatibility?  If so, why not move it
 * into the appropriate arch directory (for those architectures that
 * need it).
 */

/* If times==NULL, set access and modification to current time,
 * must be owner or have write permission.
 * Else, update from *times, must be owner or super user.
 */
asmlinkage long sys_utime(char * filename, struct utimbuf * times)
{
	int error;
	struct dentry * dentry;
	struct inode * inode;
	struct iattr newattrs;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
        enum  rsbac_target_t          rsbac_target;
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

	lock_kernel();
	dentry = namei(filename);

	error = PTR_ERR(dentry);
	if (IS_ERR(dentry))
		goto out;
	inode = dentry->d_inode;

	error = -EROFS;
	if (IS_RDONLY(inode))
		goto dput_and_out;

	/* Don't worry, the checks are done in inode_change_ok() */
	newattrs.ia_valid = ATTR_CTIME | ATTR_MTIME | ATTR_ATIME;
	if (times) {
		error = get_user(newattrs.ia_atime, &times->actime);
		if (!error) 
			error = get_user(newattrs.ia_mtime, &times->modtime);
		if (error)
			goto dput_and_out;

		newattrs.ia_valid |= ATTR_ATIME_SET | ATTR_MTIME_SET;
	} else {
		if (current->fsuid != inode->i_uid &&
		    (error = permission(inode,MAY_WRITE)) != 0)
			goto dput_and_out;
	}
	
        /* RSBAC */
        #ifdef CONFIG_RSBAC
#ifdef CONFIG_RSBAC_DEBUG
        if (rsbac_debug_aef) printk(KERN_DEBUG "sys_utime(): calling ADF\n");
#endif
        rsbac_target = T_FILE;
        if (S_ISDIR(inode->i_mode))
          rsbac_target = T_DIR;
        else if (S_ISFIFO(inode->i_mode))
          rsbac_target = T_FIFO;
        else if (S_ISLNK(inode->i_mode))
          rsbac_target = T_SYMLINK;
        rsbac_target_id.file.device = inode->i_dev;
        rsbac_target_id.file.inode  = inode->i_ino;
        rsbac_target_id.file.dentry_p = dentry;
        rsbac_attribute_value.dummy = 0;
        if (!rsbac_adf_request(R_MODIFY_ACCESS_DATA,
                               current->pid,
                               rsbac_target,
                               rsbac_target_id,
                               A_none,
                               rsbac_attribute_value))
          {
            error = -EPERM;
            goto dput_and_out;
          }
        #endif
	
	error = notify_change(dentry, &newattrs);
dput_and_out:
	dput(dentry);
out:
	unlock_kernel();
	return error;
}

#endif

/* If times==NULL, set access and modification to current time,
 * must be owner or have write permission.
 * Else, update from *times, must be owner or super user.
 */
asmlinkage long sys_utimes(char * filename, struct timeval * utimes)
{
	int error;
	struct dentry * dentry;
	struct inode * inode;
	struct iattr newattrs;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
        enum  rsbac_target_t          rsbac_target;
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

	lock_kernel();
	dentry = namei(filename);

	error = PTR_ERR(dentry);
	if (IS_ERR(dentry))
		goto out;
	inode = dentry->d_inode;

	error = -EROFS;
	if (IS_RDONLY(inode))
		goto dput_and_out;

	/* Don't worry, the checks are done in inode_change_ok() */
	newattrs.ia_valid = ATTR_CTIME | ATTR_MTIME | ATTR_ATIME;
	if (utimes) {
		struct timeval times[2];
		error = -EFAULT;
		if (copy_from_user(&times, utimes, sizeof(times)))
			goto dput_and_out;
		newattrs.ia_atime = times[0].tv_sec;
		newattrs.ia_mtime = times[1].tv_sec;
		newattrs.ia_valid |= ATTR_ATIME_SET | ATTR_MTIME_SET;
	} else {
		if ((error = permission(inode,MAY_WRITE)) != 0)
			goto dput_and_out;
	}
	
        /* RSBAC */
        #ifdef CONFIG_RSBAC
#ifdef CONFIG_RSBAC_DEBUG
        if (rsbac_debug_aef) printk(KERN_DEBUG "sys_utimes(): calling ADF\n");
#endif
        rsbac_target = T_FILE;
        if (S_ISDIR(inode->i_mode))
          rsbac_target = T_DIR;
        else if (S_ISFIFO(inode->i_mode))
          rsbac_target = T_FIFO;
        else if (S_ISLNK(inode->i_mode))
          rsbac_target = T_SYMLINK;
        rsbac_target_id.file.device = inode->i_dev;
        rsbac_target_id.file.inode  = inode->i_ino;
        rsbac_target_id.file.dentry_p = dentry;
        rsbac_attribute_value.dummy = 0;
        if (!rsbac_adf_request(R_MODIFY_ACCESS_DATA,
                               current->pid,
                               rsbac_target,
                               rsbac_target_id,
                               A_none,
                               rsbac_attribute_value))
          {
            error = -EPERM;
            goto dput_and_out;
          }
        #endif

	error = notify_change(dentry, &newattrs);
dput_and_out:
	dput(dentry);
out:
	unlock_kernel();
	return error;
}

/*
 * access() needs to use the real uid/gid, not the effective uid/gid.
 * We do this by temporarily clearing all FS-related capabilities and
 * switching the fsuid/fsgid around to the real ones.
 */
asmlinkage long sys_access(const char * filename, int mode)
{
	struct dentry * dentry;
	int old_fsuid, old_fsgid;
	kernel_cap_t old_cap;
	int res = -EINVAL;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
        enum  rsbac_target_t          rsbac_target;
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

	lock_kernel();
	if (mode != (mode & S_IRWXO))	/* where's F_OK, X_OK, W_OK, R_OK? */
		goto out;
	old_fsuid = current->fsuid;
	old_fsgid = current->fsgid;
	old_cap = current->cap_effective;

	current->fsuid = current->uid;
	current->fsgid = current->gid;

	/* Clear the capabilities if we switch to a non-root user */
	if (current->uid)
		cap_clear(current->cap_effective);
	else
		current->cap_effective = current->cap_permitted;
		
	dentry = namei(filename);
	res = PTR_ERR(dentry);
	if (!IS_ERR(dentry)) {
		res = permission(dentry->d_inode, mode);

                /* RSBAC */
                #ifdef CONFIG_RSBAC
                if(!res && dentry->d_inode)
                  {
#ifdef CONFIG_RSBAC_DEBUG
                    if (rsbac_debug_aef) printk(KERN_DEBUG "sys_access(): calling ADF\n");
#endif
                    rsbac_target = T_FILE;
                    if (S_ISDIR(dentry->d_inode->i_mode))
                      rsbac_target = T_DIR;
                    else if (S_ISFIFO(dentry->d_inode->i_mode))
                      rsbac_target = T_FIFO;
                    else if (S_ISLNK(dentry->d_inode->i_mode))
                      rsbac_target = T_SYMLINK;
                    rsbac_target_id.file.device = dentry->d_inode->i_dev;
                    rsbac_target_id.file.inode  = dentry->d_inode->i_ino;
                    rsbac_target_id.file.dentry_p = dentry;
                    rsbac_attribute_value.dummy = 0;
                    if (!rsbac_adf_request(R_GET_PERMISSIONS_DATA,
                                           current->pid,
                                           rsbac_target,
                                           rsbac_target_id,
                                           A_none,
                                           rsbac_attribute_value))
                      res = -EPERM;
                  }
                #endif

		/* SUSv2 says to return EROFS for open() and access()
		   for files on a read-only filesystem, when writing
		   is requested.  Clearly, we want to be able to run
		   a system from read-only media, so should not
		   interpret this to mean that open("/dev/tty") should
		   fail when the device node lives on a CDROM.
		   In other words, no additional check is needed here. */

		dput(dentry);
	}

	current->fsuid = old_fsuid;
	current->fsgid = old_fsgid;
	current->cap_effective = old_cap;
out:
	unlock_kernel();
	return res;
}

asmlinkage long sys_chdir(const char * filename)
{
	int error;
	struct inode *inode;
	struct dentry *dentry, *tmp;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

	lock_kernel();
	
	dentry = namei(filename);
	error = PTR_ERR(dentry);
	if (IS_ERR(dentry))
		goto out;

	inode = dentry->d_inode;

	error = -ENOTDIR;
	if (!S_ISDIR(inode->i_mode))
		goto dput_and_out;

	error = permission(inode,MAY_EXEC);
	if (error)
		goto dput_and_out;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
#ifdef CONFIG_RSBAC_DEBUG
        if (rsbac_debug_aef) printk(KERN_DEBUG "sys_chdir(): calling ADF\n");
#endif
        rsbac_target_id.dir.device = inode->i_dev;
        rsbac_target_id.dir.inode  = inode->i_ino;
        rsbac_target_id.dir.dentry_p = dentry;
        rsbac_attribute_value.dummy = 0;
        if (!rsbac_adf_request(R_CHDIR,
                               current->pid,
                               T_DIR,
                               rsbac_target_id,
                               A_none,
                               rsbac_attribute_value))
          {
            error = -EPERM;
            goto dput_and_out;
          }
        #endif
	/* exchange dentries */
	tmp = current->fs->pwd;
	current->fs->pwd = dentry;
	dentry = tmp;

dput_and_out:
	dput(dentry);
out:
	unlock_kernel();
	return error;
}

asmlinkage long sys_fchdir(unsigned int fd)
{
	struct file *file;
	struct dentry *dentry;
	struct inode *inode;
	int error;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

	lock_kernel();

	error = -EBADF;
	file = fget(fd);
	if (!file)
		goto out;

	error = -ENOENT;
	if (!(dentry = file->f_dentry))
		goto out_putf;
	if (!(inode = dentry->d_inode))
		goto out_putf;

	error = -ENOTDIR;
	if (!S_ISDIR(inode->i_mode))
		goto out_putf;

	error = permission(inode, MAY_EXEC);
	if (!error) {
		struct dentry *tmp = current->fs->pwd;

                /* RSBAC */
                #ifdef CONFIG_RSBAC
#ifdef CONFIG_RSBAC_DEBUG
                if (rsbac_debug_aef)
                  printk(KERN_DEBUG "sys_fchdir(): calling ADF\n");
#endif
                rsbac_target_id.dir.device = inode->i_dev;
                rsbac_target_id.dir.inode  = inode->i_ino;
                rsbac_target_id.dir.dentry_p = dentry;
                rsbac_attribute_value.dummy = 0;
                if (!rsbac_adf_request(R_CHDIR,
                                       current->pid,
                                       T_DIR,
                                       rsbac_target_id,
                                       A_none,
                                       rsbac_attribute_value))
                  {
                    error = -EPERM;
                    goto out_putf;
                  }
                #endif
		current->fs->pwd = dget(dentry);
		dput(tmp);
	}
out_putf:
	fput(file);
out:
	unlock_kernel();
	return error;
}

asmlinkage long sys_chroot(const char * filename)
{
	int error;
	struct inode *inode;
	struct dentry *dentry, *tmp;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

	lock_kernel();
	
	dentry = namei(filename);
	error = PTR_ERR(dentry);
	if (IS_ERR(dentry))
		goto out;

	inode = dentry->d_inode;

	error = -ENOTDIR;
	if (!S_ISDIR(inode->i_mode))
		goto dput_and_out;

	error = permission(inode,MAY_EXEC);
	if (error)
		goto dput_and_out;

	error = -EPERM;
	if (!capable(CAP_SYS_CHROOT))
		goto dput_and_out;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
#ifdef CONFIG_RSBAC_DEBUG
        if (rsbac_debug_aef)
          printk(KERN_DEBUG "sys_chroot(): calling ADF\n");
#endif
        rsbac_target_id.dir.device = inode->i_dev;
        rsbac_target_id.dir.inode  = inode->i_ino;
        rsbac_target_id.dir.dentry_p = dentry;
        rsbac_attribute_value.dummy = 0;
        if (!rsbac_adf_request(R_CHDIR,
                               current->pid,
                               T_DIR,
                               rsbac_target_id,
                               A_none,
                               rsbac_attribute_value))
          {
            error = -EPERM;
            goto dput_and_out;
          }
        #endif

	/* exchange dentries */
	tmp = current->fs->root;
	current->fs->root = dentry;
	dentry = tmp;
	error = 0;

dput_and_out:
	dput(dentry);
out:
	unlock_kernel();
	return error;
}

asmlinkage long sys_fchmod(unsigned int fd, mode_t mode)
{
	struct inode * inode;
	struct dentry * dentry;
	struct file * file;
	int err = -EBADF;
	struct iattr newattrs;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
        enum  rsbac_target_t          rsbac_target;
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

	lock_kernel();
	file = fget(fd);
	if (!file)
		goto out;

	err = -ENOENT;
	if (!(dentry = file->f_dentry))
		goto out_putf;
	if (!(inode = dentry->d_inode))
		goto out_putf;

	err = -EROFS;
	if (IS_RDONLY(inode))
		goto out_putf;
	err = -EPERM;
	if (IS_IMMUTABLE(inode) || IS_APPEND(inode))
		goto out_putf;
        /* RSBAC */
        #ifdef CONFIG_RSBAC
#ifdef CONFIG_RSBAC_DEBUG
        if (rsbac_debug_aef) printk(KERN_DEBUG "sys_fchmod(): calling ADF\n");
#endif
        rsbac_target = T_FILE;
        if (S_ISDIR(inode->i_mode))
          rsbac_target = T_DIR;
        else if (S_ISFIFO(inode->i_mode))
          rsbac_target = T_FIFO;
        else if (S_ISLNK(inode->i_mode))
          rsbac_target = T_SYMLINK;
        rsbac_target_id.file.device = inode->i_dev;
        rsbac_target_id.file.inode  = inode->i_ino;
        rsbac_target_id.file.dentry_p = dentry;
        rsbac_attribute_value.dummy = 0;
        if (!rsbac_adf_request(R_MODIFY_PERMISSIONS_DATA,
                               current->pid,
                               rsbac_target,
                               rsbac_target_id,
                               A_none,
                               rsbac_attribute_value))
          {
            err = -EPERM;
            goto out_putf;
          }
        #endif
        
	if (mode == (mode_t) -1)
		mode = inode->i_mode;
	newattrs.ia_mode = (mode & S_IALLUGO) | (inode->i_mode & ~S_IALLUGO);
	newattrs.ia_valid = ATTR_MODE | ATTR_CTIME;
	err = notify_change(dentry, &newattrs);

out_putf:
	fput(file);
out:
	unlock_kernel();
	return err;
}

asmlinkage long sys_chmod(const char * filename, mode_t mode)
{
	struct dentry * dentry;
	struct inode * inode;
	int error;
	struct iattr newattrs;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
        enum  rsbac_target_t          rsbac_target;
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

	lock_kernel();
	dentry = namei(filename);

	error = PTR_ERR(dentry);
	if (IS_ERR(dentry))
		goto out;
	inode = dentry->d_inode;

	error = -EROFS;
	if (IS_RDONLY(inode))
		goto dput_and_out;

	error = -EPERM;
	if (IS_IMMUTABLE(inode) || IS_APPEND(inode))
		goto dput_and_out;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
#ifdef CONFIG_RSBAC_DEBUG
        if (rsbac_debug_aef) printk(KERN_DEBUG "sys_chmod(): calling ADF\n");
#endif
        rsbac_target = T_FILE;
        if (S_ISDIR(inode->i_mode))
          rsbac_target = T_DIR;
        else if (S_ISFIFO(inode->i_mode))
          rsbac_target = T_FIFO;
        else if (S_ISLNK(inode->i_mode))
          rsbac_target = T_SYMLINK;
        rsbac_target_id.file.device = inode->i_dev;
        rsbac_target_id.file.inode  = inode->i_ino;
        rsbac_target_id.file.dentry_p = dentry;
        rsbac_attribute_value.dummy = 0;
        if (!rsbac_adf_request(R_MODIFY_PERMISSIONS_DATA,
                               current->pid,
                               rsbac_target,
                               rsbac_target_id,
                               0,
                               rsbac_attribute_value))
          {
            error = -EPERM;
            goto dput_and_out;
          }
        #endif

	if (mode == (mode_t) -1)
		mode = inode->i_mode;
	newattrs.ia_mode = (mode & S_IALLUGO) | (inode->i_mode & ~S_IALLUGO);
	newattrs.ia_valid = ATTR_MODE | ATTR_CTIME;
	error = notify_change(dentry, &newattrs);

dput_and_out:
	dput(dentry);
out:
	unlock_kernel();
	return error;
}

static int chown_common(struct dentry * dentry, uid_t user, gid_t group)
{
	struct inode * inode;
	int error;
	struct iattr newattrs;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
        enum  rsbac_target_t          rsbac_target;
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

	error = -ENOENT;
	if (!(inode = dentry->d_inode)) {
		printk("chown_common: NULL inode\n");
		goto out;
	}
	error = -EROFS;
	if (IS_RDONLY(inode))
		goto out;
	error = -EPERM;
	if (IS_IMMUTABLE(inode) || IS_APPEND(inode))
		goto out;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
#ifdef CONFIG_RSBAC_DEBUG
        if (rsbac_debug_aef) printk(KERN_DEBUG "chown_common() [sys_*chown]: calling ADF\n");
#endif
        rsbac_target = T_FILE;
        if (S_ISDIR(inode->i_mode))
          rsbac_target = T_DIR;
        else if (S_ISFIFO(inode->i_mode))
          rsbac_target = T_FIFO;
        else if (S_ISLNK(inode->i_mode))
          rsbac_target = T_SYMLINK;
        rsbac_target_id.file.device = inode->i_dev;
        rsbac_target_id.file.inode  = inode->i_ino;
        rsbac_target_id.file.dentry_p = dentry;
        rsbac_attribute_value.dummy = 0;
        if (!rsbac_adf_request(R_CHANGE_OWNER,
                               current->pid,
                               rsbac_target,
                               rsbac_target_id,
                               A_none,
                               rsbac_attribute_value))
          {
            error = -EPERM;
            goto out;
          }
        #endif

	if (user == (uid_t) -1)
		user = inode->i_uid;
	if (group == (gid_t) -1)
		group = inode->i_gid;
	newattrs.ia_mode = inode->i_mode;
	newattrs.ia_uid = user;
	newattrs.ia_gid = group;
	newattrs.ia_valid =  ATTR_UID | ATTR_GID | ATTR_CTIME;
	/*
	 * If the user or group of a non-directory has been changed by a
	 * non-root user, remove the setuid bit.
	 * 19981026	David C Niemi <niemi@tux.org>
	 *
	 * Changed this to apply to all users, including root, to avoid
	 * some races. This is the behavior we had in 2.0. The check for
	 * non-root was definitely wrong for 2.2 anyway, as it should
	 * have been using CAP_FSETID rather than fsuid -- 19990830 SD.
	 */
	if ((inode->i_mode & S_ISUID) == S_ISUID &&
		!S_ISDIR(inode->i_mode))
	{
		newattrs.ia_mode &= ~S_ISUID;
		newattrs.ia_valid |= ATTR_MODE;
	}
	/*
	 * Likewise, if the user or group of a non-directory has been changed
	 * by a non-root user, remove the setgid bit UNLESS there is no group
	 * execute bit (this would be a file marked for mandatory locking).
	 * 19981026	David C Niemi <niemi@tux.org>
	 *
	 * Removed the fsuid check (see the comment above) -- 19990830 SD.
	 */
	if (((inode->i_mode & (S_ISGID | S_IXGRP)) == (S_ISGID | S_IXGRP)) 
		&& !S_ISDIR(inode->i_mode))
	{
		newattrs.ia_mode &= ~S_ISGID;
		newattrs.ia_valid |= ATTR_MODE;
	}
	error = DQUOT_TRANSFER(dentry, &newattrs);
out:
	return error;
}

asmlinkage long sys_chown(const char * filename, uid_t user, gid_t group)
{
	struct dentry * dentry;
	int error;

	lock_kernel();
	dentry = namei(filename);

	error = PTR_ERR(dentry);
	if (!IS_ERR(dentry)) {
		error = chown_common(dentry, user, group);
		dput(dentry);
	}
	unlock_kernel();
	return error;
}

asmlinkage long sys_lchown(const char * filename, uid_t user, gid_t group)
{
	struct dentry * dentry;
	int error;

	lock_kernel();
	dentry = lnamei(filename);

	error = PTR_ERR(dentry);
	if (!IS_ERR(dentry)) {
		error = chown_common(dentry, user, group);
		dput(dentry);
	}
	unlock_kernel();
	return error;
}


asmlinkage long sys_fchown(unsigned int fd, uid_t user, gid_t group)
{
	struct dentry * dentry;
	struct file * file;
	int error = -EBADF;

	lock_kernel();
	file = fget(fd);
	if (!file)
		goto out;
	error = -ENOENT;
	if ((dentry = file->f_dentry) != NULL)
		error = chown_common(dentry, user, group);
	fput(file);

out:
	unlock_kernel();
	return error;
}

/*
 * Note that while the flag value (low two bits) for sys_open means:
 *	00 - read-only
 *	01 - write-only
 *	10 - read-write
 *	11 - special
 * it is changed into
 *	00 - no permissions needed
 *	01 - read-permission
 *	10 - write-permission
 *	11 - read-write
 * for the internal routines (ie open_namei()/follow_link() etc). 00 is
 * used by symlinks.
 */
struct file *filp_open(const char * filename, int flags, int mode)
{
	struct inode * inode;
	struct dentry * dentry;
	struct file * f;
	int flag,error;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
        enum  rsbac_target_t          rsbac_target;
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_target_id_t       rsbac_new_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        enum  rsbac_adf_request_t     rsbac_adf_req;
        #endif

	error = -ENFILE;
	f = get_empty_filp();
	if (!f)
		goto out;
#if BITS_PER_LONG != 32
	flags |= O_LARGEFILE;
#endif
	f->f_flags = flag = flags;
	f->f_mode = (flag+1) & O_ACCMODE;
	if (f->f_mode)
		flag++;
	if (flag & O_TRUNC)
		flag |= 2;
	dentry = open_namei(filename,flag,mode);
	error = PTR_ERR(dentry);
	if (IS_ERR(dentry))
		goto cleanup_file;
	inode = dentry->d_inode;
	if (f->f_mode & FMODE_WRITE) {
		error = get_write_access(inode);
		if (error)
			goto cleanup_dentry;
	}

	f->f_dentry = dentry;
	f->f_pos = 0;
	f->f_reada = 0;
	f->f_op = NULL;
	if (inode->i_shadow && inode->i_shadow->i_shadow_op) {
		error = inode->i_shadow->i_shadow_op->open_file(inode->i_shadow, f);
		if (error)
			goto cleanup_all;
	}
	if (inode->i_op)
		f->f_op = inode->i_op->default_file_ops;
	if (f->f_op && f->f_op->open) {
		error = f->f_op->open(inode,f);
		if (error)
			goto cleanup_all;
	}
	f->f_flags &= ~(O_CREAT | O_EXCL | O_NOCTTY | O_TRUNC);

        /* RSBAC: notify adf of opened file */
        #ifdef CONFIG_RSBAC
        /* get target type and id clear */
        rsbac_target = T_FILE;
        if(   S_ISBLK(inode->i_mode)
           || S_ISCHR(inode->i_mode))
          {
            rsbac_target = T_DEV;
            if(S_ISBLK(inode->i_mode))
              {
                rsbac_target_id.dev.type = D_block;
              }
            else
              {
                rsbac_target_id.dev.type = D_char;
              }
            rsbac_target_id.dev.id = inode->i_rdev;
          }
        else /* must be file, fifo or dir */
          {
            if(S_ISDIR(inode->i_mode))
              rsbac_target = T_DIR;
            else if (S_ISFIFO(inode->i_mode))
              rsbac_target = T_FIFO;
            rsbac_target_id.file.device = inode->i_dev;
            rsbac_target_id.file.inode  = inode->i_ino;
            rsbac_target_id.file.dentry_p = dentry;
          }
        /* determine request type */
        rsbac_adf_req = R_NONE;
        if (flag & O_APPEND)
          rsbac_adf_req = R_APPEND_OPEN;
        else
          if ((flag & FMODE_WRITE) && (flag & FMODE_READ))
            rsbac_adf_req = R_READ_WRITE_OPEN;
          else
            if (flag & FMODE_WRITE)
              rsbac_adf_req = R_WRITE_OPEN;
            else
              if (flag & FMODE_READ)
                {
                  if(rsbac_target == T_DIR)
                    rsbac_adf_req = R_READ;
                  else
                    rsbac_adf_req = R_READ_OPEN;
                }
        if (rsbac_adf_req != R_NONE)
          {
            rsbac_new_target_id.dummy = 0;
            rsbac_attribute_value.dummy = 0;
            if (rsbac_adf_set_attr(rsbac_adf_req,
                                   current->pid,
                                   rsbac_target,
                                   rsbac_target_id,
                                   T_NONE,
                                   rsbac_new_target_id,
                                   A_none,
                                   rsbac_attribute_value))
              {
                printk(KERN_WARNING
                       "filp_open() [do_open() [sys_open()]]: rsbac_adf_set_attr() returned error\n");
              }
          }
        #endif

	return f;

cleanup_all:
	if (f->f_mode & FMODE_WRITE)
		put_write_access(inode);
cleanup_dentry:
	f->f_dentry = NULL;
	dput(dentry);
cleanup_file:
	put_filp(f);
out:
	return ERR_PTR(error);
}

/*
 * Find an empty file descriptor entry, and mark it busy.
 */
int get_unused_fd(void)
{
	struct files_struct * files = current->files;
	int fd, error;
	
repeat:
	error = -EMFILE;

	fd = find_next_zero_bit(files->open_fds, 
				current->files->max_fdset, 
				files->next_fd);
	/*
	 * N.B. For clone tasks sharing a files structure, this test
	 * will limit the total number of files that can be opened.
	 */
	if (fd >= current->rlim[RLIMIT_NOFILE].rlim_cur)
		goto out;

	/* Do we need to expand the fdset array? */
	if (fd >= current->files->max_fdset) {
		error = expand_fdset(files, fd + 1);
		if (!error)
			goto repeat;
		goto out;
	}
	
	/* 
	 * Check whether we need to expand the fd array.
	 */
	if (fd >= files->max_fds) {
		error = expand_fd_array(files, fd + 1);
		if (!error)
			goto repeat;
		goto out;
	}

	FD_SET(fd, files->open_fds);
	FD_CLR(fd, files->close_on_exec);
	files->next_fd = fd + 1;
#if 1
	/* Sanity check */
	if (files->fd[fd] != NULL) {
		printk("get_unused_fd: slot %d not NULL!\n", fd);
		files->fd[fd] = NULL;
	}
#endif
	error = fd;

out:
#ifdef FDSET_DEBUG	
	if (error < 0)
		printk (KERN_ERR __FUNCTION__ ": return %d\n", error);
#endif
	return error;
}

asmlinkage int sys_open(const char * filename, int flags, int mode)
{
	char * tmp;
	int fd, error;

	tmp = getname(filename);
	fd = PTR_ERR(tmp);
	if (!IS_ERR(tmp)) {
		lock_kernel();
		fd = get_unused_fd();
		if (fd >= 0) {
			struct file * f = filp_open(tmp, flags, mode);
			error = PTR_ERR(f);
			if (IS_ERR(f))
				goto out_error;
			fd_install(fd, f);
		}
out:
		unlock_kernel();
		putname(tmp);
	}
	return fd;

out_error:
	put_unused_fd(fd);
	fd = error;
	goto out;
}

#ifndef __alpha__

/*
 * For backward compatibility?  Maybe this should be moved
 * into arch/i386 instead?
 */
asmlinkage long sys_creat(const char * pathname, int mode)
{
	return sys_open(pathname, O_CREAT | O_WRONLY | O_TRUNC, mode);
}

#endif

/*
 * Called when retiring the last use of a file pointer.
 */
void __fput(struct file *filp)
{
	struct dentry * dentry = filp->f_dentry;
	struct inode * inode = dentry->d_inode;

	if (filp->f_op && filp->f_op->release)
		filp->f_op->release(inode, filp);
	if (inode->i_shadow && inode->i_shadow->i_shadow_op)
		inode->i_shadow->i_shadow_op->release_file(inode->i_shadow, filp);
	filp->f_dentry = NULL;
	if (filp->f_mode & FMODE_WRITE)
		put_write_access(inode);
	dput(dentry);
}

/*
 * "id" is the POSIX thread ID. We use the
 * files pointer for this..
 */
int filp_close(struct file *filp, fl_owner_t id)
{
	int retval;
	struct dentry *dentry = filp->f_dentry;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
        enum  rsbac_target_t          rsbac_target = T_NONE;
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_target_id_t       rsbac_new_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

	if (filp->f_count == 0) {
		printk("VFS: Close: file count is 0\n");
		return 0;
	}
        /* RSBAC: calling adf */
        #ifdef CONFIG_RSBAC
        if (   dentry
            && dentry->d_inode
            && !S_ISSOCK(dentry->d_inode->i_mode))
          {
#ifdef CONFIG_RSBAC_DEBUG
            if (rsbac_debug_aef)
              printk(KERN_DEBUG "close_fp() [sys_close]: calling ADF\n");
#endif
            rsbac_target = T_FILE;
            if(   S_ISBLK(dentry->d_inode->i_mode)
               || S_ISCHR(dentry->d_inode->i_mode))
              {
                rsbac_target = T_DEV;
                if(S_ISBLK(dentry->d_inode->i_mode))
                  {
                    rsbac_target_id.dev.type = D_block;
                  }
                else
                  {
                    rsbac_target_id.dev.type = D_char;
                  }
                rsbac_target_id.dev.id = dentry->d_inode->i_rdev;
              }
            else /* must be file or dir */
              {
                if (S_ISDIR(dentry->d_inode->i_mode))
                  rsbac_target = T_DIR;
                else if (S_ISFIFO(dentry->d_inode->i_mode))
                  rsbac_target = T_FIFO;
                rsbac_target_id.file.device = dentry->d_inode->i_dev;
                rsbac_target_id.file.inode  = dentry->d_inode->i_ino;
                rsbac_target_id.file.dentry_p = dentry;
              }
            rsbac_attribute_value.dummy = 0;
            if (!rsbac_adf_request(R_CLOSE,
                                   current->pid,
                                   rsbac_target,
                                   rsbac_target_id,
                                   A_none,
                                   rsbac_attribute_value))
              {
                printk(KERN_WARNING "close_fp() [sys_close]: ADF-call returned NOT_GRANTED\n");
                return -EPERM;
              }
          }
        #endif

	retval = 0;
	if (filp->f_op && filp->f_op->flush)
		retval = filp->f_op->flush(filp);
	if (dentry->d_inode)
		locks_remove_posix(filp, id);
	fput(filp);

        /* RSBAC: notifying adf */
        #ifdef CONFIG_RSBAC
        if (   dentry
            && dentry->d_inode
            && !S_ISSOCK(dentry->d_inode->i_mode))
          {
#ifdef CONFIG_RSBAC_DEBUG
            if (rsbac_debug_aef)
              printk(KERN_DEBUG "close_fp() [sys_close]: notifying ADF\n");
#endif
            rsbac_new_target_id.dummy = 0;
            rsbac_attribute_value.dummy = 0;
            if (rsbac_adf_set_attr(R_CLOSE,
                                   current->pid,
                                   rsbac_target,
                                   rsbac_target_id,
                                   T_NONE,
                                   rsbac_new_target_id,
                                   A_none,
                                   rsbac_attribute_value))
              {
                printk(KERN_WARNING
                       "close_fp() [sys_close]: rsbac_adf_set_attr() returned error\n");
              }
          }
        #endif

	return retval;
}

/*
 * Careful here! We test whether the file pointer is NULL before
 * releasing the fd. This ensures that one clone task can't release
 * an fd while another clone is opening it.
 */
asmlinkage int sys_close(unsigned int fd)
{
	int error;
	struct file * filp;

        /* RSBAC */
        #ifdef CONFIG_RSBAC
        int                           rsbac_is_socket = 0;
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_target_id_t       rsbac_new_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

	lock_kernel();
	error = -EBADF;
	filp = fcheck(fd);
	if (filp) {
		struct files_struct * files = current->files;

                /* RSBAC: remember, whether it was a socket, and request ADF
                 * This is special, because closing a socket means deleting it.
                 * The real close might have been done via sys_shutdown, then
                 * here is nothing more to do.
                 */
                #ifdef CONFIG_RSBAC
                if (   S_ISSOCK(filp->f_dentry->d_inode->i_mode)
                    && filp->f_dentry->d_inode->u.socket_i.sk
                   )
                  {
                    rsbac_is_socket = 1;
                    rsbac_target_id.ipc.type  = I_sock;
                    rsbac_target_id.ipc.id.sock
                      = filp->f_dentry->d_inode->u.socket_i.sk;
#ifdef CONFIG_RSBAC_DEBUG
                    if (rsbac_debug_aef)
                      printk(KERN_DEBUG "sys_close: calling ADF for DELETE socket\n");
#endif
                    rsbac_attribute_value.dummy = 0;
                    if (!rsbac_adf_request(R_DELETE,
                                           current->pid,
                                           T_IPC,
                                           rsbac_target_id,
                                           A_none,
                                           rsbac_attribute_value))
                      {
                        printk(KERN_WARNING
                               "sys_close(): ADF-call for DELETE on socket returned NOT_GRANTED\n");
                        unlock_kernel();
                        return -EPERM;
                      }
                  }
                #endif

		files->fd[fd] = NULL;
		put_unused_fd(fd);
		FD_CLR(fd, files->close_on_exec);
 		error = filp_close(filp, files);

                /* RSBAC: notifying adf / socket */
                #ifdef CONFIG_RSBAC
                if (!error && rsbac_is_socket)
                  {
#ifdef CONFIG_RSBAC_DEBUG
                    if (rsbac_debug_aef)
                      printk(KERN_DEBUG
                        "sys_close: calling ADF for DELETE socket\n");
#endif
                    rsbac_new_target_id.dummy = 0;
                    rsbac_attribute_value.dummy = 0;
                    if (rsbac_adf_set_attr(R_DELETE,
                                           current->pid,
                                           T_IPC,
                                           rsbac_target_id,
                                           T_NONE,
                                           rsbac_new_target_id,
                                           A_none,
                                           rsbac_attribute_value))
                      {
                        printk(KERN_WARNING 
                               "sys_close(): rsbac_adf_set_attr() for DELETE on socket returned error\n");
                      }
                  }
                #endif
	}
	unlock_kernel();
	return error;
}

/*
 * This routine simulates a hangup on the tty, to arrange that users
 * are given clean terminals at login time.
 */
asmlinkage long sys_vhangup(void)
{
	int ret = -EPERM;

	lock_kernel();
	if (!capable(CAP_SYS_TTY_CONFIG))
		goto out;
	/* If there is a controlling tty, hang it up */
	if (current->tty)
		tty_vhangup(current->tty);
	ret = 0;
out:
	unlock_kernel();
	return ret;
}
