/*
 *  linux/fs/read_write.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 */

#include <linux/malloc.h> 
#include <linux/stat.h>
#include <linux/fcntl.h>
#include <linux/file.h>
#include <linux/uio.h>
#include <linux/smp_lock.h>

#include <asm/uaccess.h>

/* RSBAC */
#ifdef CONFIG_RSBAC_RW
#include <rsbac/adf.h>
#endif

static loff_t default_llseek(struct file *file, loff_t offset, int origin)
{
	long long retval;

	switch (origin) {
		case 2:
			offset += file->f_dentry->d_inode->i_size;
			break;
		case 1:
			offset += file->f_pos;
	}
	retval = -EINVAL;
	if (offset >= 0) {
		if (offset != file->f_pos) {
			file->f_pos = offset;
			file->f_reada = 0;
			file->f_version = ++global_event;
		}
		retval = offset;
	}
	return retval;
}

static inline loff_t llseek(struct file *file, loff_t offset, int origin)
{
	loff_t (*fn)(struct file *, loff_t, int);

	fn = default_llseek;
	if (file->f_op && file->f_op->llseek)
		fn = file->f_op->llseek;
	return fn(file, offset, origin);
}

off_t do_lseek(struct file *file, off_t offset, unsigned int origin)
{
	off_t retval;
        off_t oldpos;
	struct dentry * dentry;
	struct inode * inode;

	/* N.B. Shouldn't this be ENOENT?? */
	if (!(dentry = file->f_dentry) || !(inode = dentry->d_inode))
		return -EBADF;
        oldpos = file->f_pos;
	retval = -EINVAL;
	if (origin <= 2)
		retval = llseek(file, offset, origin);

	/* Demand L-F-S compliance only from normal files,
           thus raw devices can do whatever they please.. */
        if (!(file->f_flags & O_LARGEFILE) &&
            retval >= 0 && S_ISREG(inode->i_mode) &&
            file->f_pos > 0x7fffffff) {
                file->f_pos = oldpos;
                retval = -EOVERFLOW;
        }
        
	return retval;
}

asmlinkage off_t sys_lseek(unsigned int fd, off_t offset, unsigned int origin)
{
	off_t retval;
	struct file * file;

	lock_kernel();
	retval = -EBADF;
	file = fget(fd);
	if (file)
	{
		retval = do_lseek(file, offset, origin);
		fput(file);
	}
	unlock_kernel();
	return retval;
}

#if !defined(__alpha__)
asmlinkage int sys_llseek(unsigned int fd, unsigned long offset_high,
			  unsigned long offset_low, loff_t * result,
			  unsigned int origin)
{
	int retval;
	struct file * file;
	struct dentry * dentry;
	struct inode * inode;
	loff_t offset, oldpos;

	lock_kernel();
	retval = -EBADF;
	file = fget(fd);
	if (!file)
		goto bad;
	/* N.B. Shouldn't this be ENOENT?? */
	if (!(dentry = file->f_dentry) ||
	    !(inode = dentry->d_inode))
		goto out_putf;
	retval = -EINVAL;
	if (origin > 2)
		goto out_putf;

	oldpos = file->f_pos;
	offset = llseek(file, ((loff_t) offset_high << 32) | offset_low,
			origin);

	retval = (int)offset;
	if (offset >= 0) {
		retval = -EFAULT;
		if (!copy_to_user(result, &offset, sizeof(offset)))
			retval = 0;
	}
	if (!(file->f_flags & O_LARGEFILE) && S_ISREG(inode->i_mode) &&
	    file->f_pos > 0x7fffffff) {
		/* The target position isn't presentable without
		   O_LARGEFILE flag being set --> yield error, and
		   restore the file position. */
		file->f_pos = oldpos;
		retval = -EOVERFLOW;
	}
out_putf:
	fput(file);
bad:
	unlock_kernel();
	return retval;
}
#endif

asmlinkage ssize_t sys_read(unsigned int fd, char * buf, size_t count)
{
	ssize_t ret;
	struct file * file;
	ssize_t (*read)(struct file *, char *, size_t, loff_t *);

        /* RSBAC */
        #ifdef CONFIG_RSBAC_RW
        enum  rsbac_target_t          rsbac_target = T_NONE;
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_target_id_t       rsbac_new_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

	lock_kernel();

	ret = -EBADF;
	file = fget(fd);
	if (!file)
		goto bad_file;
	if (!(file->f_mode & FMODE_READ))
		goto out;
	ret = locks_verify_area(FLOCK_VERIFY_READ, file->f_dentry->d_inode,
				file, file->f_pos, count);
	if (ret)
		goto out;
	ret = -EINVAL;
	if (!file->f_op || !(read = file->f_op->read))
		goto out;

        /* RSBAC */
        #ifdef CONFIG_RSBAC_RW
#ifdef CONFIG_RSBAC_DEBUG
        if (rsbac_debug_aef)
          printk(KERN_DEBUG "sys_read(): calling ADF\n");
#endif
        /* get target type and id clear */
        #ifdef CONFIG_RSBAC_RW_SOCK
        if(S_ISSOCK(file->f_dentry->d_inode->i_mode))
          {
            rsbac_target = T_IPC;
            rsbac_target_id.ipc.type = I_sock;
            rsbac_target_id.ipc.id.sock = file->f_dentry->d_inode->u.socket_i.sk;
          }
        else
        #endif
        if(   S_ISFIFO(file->f_dentry->d_inode->i_mode)
              /* named FIFO only */
           && file->f_dentry->d_inode->i_dev
          )
          {
            rsbac_target = T_FIFO;
            rsbac_target_id.fifo.device = file->f_dentry->d_inode->i_dev;
            rsbac_target_id.fifo.inode  = file->f_dentry->d_inode->i_ino;
            rsbac_target_id.fifo.dentry_p = file->f_dentry;
          }
        else
        if(S_ISREG(file->f_dentry->d_inode->i_mode))
          {
            /* SHM special files */
            if (   file->f_dentry->d_inode->i_sb
                && (file->f_dentry->d_inode->i_sb->s_magic == SHM_FS_MAGIC)
               )
              {
                rsbac_target = T_IPC;
                rsbac_target_id.ipc.type = I_shm;
                rsbac_target_id.ipc.id.id_nr = file->f_dentry->d_inode->i_ino;
              }
            else
              {
                rsbac_target = T_FILE;
                rsbac_target_id.file.device = file->f_dentry->d_inode->i_dev;
                rsbac_target_id.file.inode  = file->f_dentry->d_inode->i_ino;
                rsbac_target_id.file.dentry_p = file->f_dentry;
              }
          }
        else
        if(S_ISBLK(file->f_dentry->d_inode->i_mode))
          {
            rsbac_target = T_DEV;
            rsbac_target_id.dev.type = D_block;
            rsbac_target_id.dev.id = file->f_dentry->d_inode->i_rdev;
          }
        else
        if(S_ISCHR(file->f_dentry->d_inode->i_mode))
          {
            rsbac_target = T_DEV;
            rsbac_target_id.dev.type = D_char;
            rsbac_target_id.dev.id = file->f_dentry->d_inode->i_rdev;
          }
        rsbac_attribute_value.dummy = 0;
        if (   (rsbac_target != T_NONE)
            && !rsbac_adf_request(R_READ,
                                  current->pid,
                                  rsbac_target,
                                  rsbac_target_id,
                                  A_none,
                                  rsbac_attribute_value)
           )
          {
            ret = -EPERM;
            goto out;
          }
        #endif /* CONFIG_RSBAC_RW */

	ret = read(file, buf, count, &file->f_pos);

        /* RSBAC: notify adf of read file etc. */
        #ifdef CONFIG_RSBAC_RW
        if (   (ret > 0)
            && (rsbac_target != T_NONE)
           )
          {
            rsbac_new_target_id.dummy = 0;
            if (rsbac_adf_set_attr(R_READ,
                                   current->pid,
                                   rsbac_target,
                                   rsbac_target_id,
                                   T_NONE,
                                   rsbac_new_target_id,
                                   A_none,
                                   rsbac_attribute_value))
              {
                printk(KERN_WARNING
                       "sys_read: rsbac_adf_set_attr() returned error\n");
              }
          }
        #endif

out:
	fput(file);
bad_file:
	unlock_kernel();
	return ret;
}

asmlinkage ssize_t sys_write(unsigned int fd, const char * buf, size_t count)
{
	ssize_t ret;
	struct file * file;
	struct inode * inode;
	ssize_t (*write)(struct file *, const char *, size_t, loff_t *);

        /* RSBAC */
        #ifdef CONFIG_RSBAC_RW
        enum  rsbac_target_t          rsbac_target = T_NONE;
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_target_id_t       rsbac_new_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

	lock_kernel();

	ret = -EBADF;
	file = fget(fd);
	if (!file)
		goto bad_file;
	if (!(file->f_mode & FMODE_WRITE))
		goto out;
	inode = file->f_dentry->d_inode;
	ret = locks_verify_area(FLOCK_VERIFY_WRITE, inode, file,
				file->f_pos, count);
	if (ret)
		goto out;
	ret = -EINVAL;
	if (!file->f_op || !(write = file->f_op->write))
		goto out;

        /* RSBAC */
        #ifdef CONFIG_RSBAC_RW
#ifdef CONFIG_RSBAC_DEBUG
        if (rsbac_debug_aef)
          printk(KERN_DEBUG "sys_write(): calling ADF\n");
#endif
        /* get target type and id clear */
        #ifdef CONFIG_RSBAC_RW_SOCK
        if(S_ISSOCK(inode->i_mode))
          {
            rsbac_target = T_IPC;
            rsbac_target_id.ipc.type = I_sock;
            rsbac_target_id.ipc.id.sock = inode->u.socket_i.sk;
          }
        else
        #endif
        if(   S_ISFIFO(inode->i_mode)
              /* named FIFO only */
           && inode->i_dev
          )
          {
            rsbac_target = T_FIFO;
            rsbac_target_id.fifo.device = inode->i_dev;
            rsbac_target_id.fifo.inode  = inode->i_ino;
            rsbac_target_id.fifo.dentry_p = file->f_dentry;
          }
        else
        if(S_ISREG(inode->i_mode))
          {
            /* SHM special files */
            if (   inode->i_sb
                && (inode->i_sb->s_magic == SHM_FS_MAGIC)
               )
              {
                rsbac_target = T_IPC;
                rsbac_target_id.ipc.type = I_shm;
                rsbac_target_id.ipc.id.id_nr = inode->i_ino;
              }
            else
              {
                rsbac_target = T_FILE;
                rsbac_target_id.file.device = inode->i_dev;
                rsbac_target_id.file.inode  = inode->i_ino;
                rsbac_target_id.file.dentry_p = file->f_dentry;
              }
          }
        else
        if(S_ISBLK(inode->i_mode))
          {
            rsbac_target = T_DEV;
            rsbac_target_id.dev.type = D_block;
            rsbac_target_id.dev.id = inode->i_rdev;
          }
        else
        if(S_ISCHR(inode->i_mode))
          {
            rsbac_target = T_DEV;
            rsbac_target_id.dev.type = D_char;
            rsbac_target_id.dev.id = inode->i_rdev;
          }
        rsbac_attribute_value.dummy = 0;
        if (   (rsbac_target != T_NONE)
            && !rsbac_adf_request(R_WRITE,
                                  current->pid,
                                  rsbac_target,
                                  rsbac_target_id,
                                  A_none,
                                  rsbac_attribute_value)
           )
          {
            ret = -EPERM;
            goto out;
          }
        #endif /* CONFIG_RSBAC_RW */

	fs_down(&inode->i_sem);
	ret = write(file, buf, count, &file->f_pos);
	fs_up(&inode->i_sem);

        /* RSBAC: notify adf of read file etc. */
        #ifdef CONFIG_RSBAC_RW
        if (   (ret > 0)
            && (rsbac_target != T_NONE)
           )
          {
            rsbac_new_target_id.dummy = 0;
            if (rsbac_adf_set_attr(R_WRITE,
                                   current->pid,
                                   rsbac_target,
                                   rsbac_target_id,
                                   T_NONE,
                                   rsbac_new_target_id,
                                   A_none,
                                   rsbac_attribute_value))
              {
                printk(KERN_WARNING
                       "sys_write: rsbac_adf_set_attr() returned error\n");
              }
          }
        #endif

out:
	fput(file);
bad_file:
	unlock_kernel();
	return ret;
}


static ssize_t do_readv_writev(int type, struct file *file,
			       const struct iovec * vector,
			       unsigned long count)
{
	typedef ssize_t (*io_fn_t)(struct file *, char *, size_t, loff_t *);

	size_t tot_len;
	struct iovec iovstack[UIO_FASTIOV];
	struct iovec *iov=iovstack;
	ssize_t ret, i;
	io_fn_t fn;
	struct inode *inode;

        /* RSBAC */
        #ifdef CONFIG_RSBAC_RW
        enum  rsbac_adf_request_t     rsbac_request;
        enum  rsbac_target_t          rsbac_target = T_NONE;
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_target_id_t       rsbac_new_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

	/*
	 * First get the "struct iovec" from user memory and
	 * verify all the pointers
	 */
	ret = 0;
	if (!count)
		goto out_nofree;
	ret = -EINVAL;
	if (count > UIO_MAXIOV)
		goto out_nofree;
	if (count > UIO_FASTIOV) {
		ret = -ENOMEM;
		iov = kmalloc(count*sizeof(struct iovec), GFP_KERNEL);
		if (!iov)
			goto out_nofree;
	}
	ret = -EFAULT;
	if (copy_from_user(iov, vector, count*sizeof(*vector)))
		goto out;

	/* BSD readv/writev returns EINVAL if one of the iov_len
	   values < 0 or tot_len overflowed a 32-bit integer. -ink */
	tot_len = 0;
	ret = -EINVAL;
	for (i = 0 ; i < count ; i++) {
		size_t tmp = tot_len;
		int len = iov[i].iov_len;
		if (len < 0)
			goto out;
		(u32)tot_len += len;
		if (tot_len < tmp || tot_len < (u32)len)
			goto out;
	}

	inode = file->f_dentry->d_inode;
	/* VERIFY_WRITE actually means a read, as we write to user space */
	ret = locks_verify_area((type == VERIFY_WRITE
				 ? FLOCK_VERIFY_READ : FLOCK_VERIFY_WRITE),
				inode, file, file->f_pos, tot_len);
	if (ret) goto out;

	/*
	 * Then do the actual IO.  Note that sockets need to be handled
	 * specially as they have atomicity guarantees and can handle
	 * iovec's natively
	 */
	if (inode->i_sock) {
		ret = sock_readv_writev(type,inode,file,iov,count,tot_len);
		goto out;
	}

	ret = -EINVAL;
	if (!file->f_op)
		goto out;

        /* RSBAC */
        #ifdef CONFIG_RSBAC_RW
#ifdef CONFIG_RSBAC_DEBUG
        if (rsbac_debug_aef)
          printk(KERN_DEBUG "do_readv_writev() [sys_vread(), sys_vwrite()]: calling ADF\n");
#endif
        /* get target type and id clear */
        #ifdef CONFIG_RSBAC_RW_SOCK
        if(S_ISSOCK(inode->i_mode))
          {
            rsbac_target = T_IPC;
            rsbac_target_id.ipc.type = I_sock;
            rsbac_target_id.ipc.id.sock = inode->u.socket_i.sk;
          }
        else
        #endif
        if(   S_ISFIFO(inode->i_mode)
           /* named FIFO only */
           && inode->i_dev
          )
          {
            rsbac_target = T_FIFO;
            rsbac_target_id.fifo.device = inode->i_dev;
            rsbac_target_id.fifo.inode  = inode->i_ino;
            rsbac_target_id.fifo.dentry_p = file->f_dentry;
          }
        else
        if(S_ISREG(inode->i_mode))
          {
            /* SHM special files */
            if (   inode->i_sb
                && (inode->i_sb->s_magic == SHM_FS_MAGIC)
               )
              {
                rsbac_target = T_IPC;
                rsbac_target_id.ipc.type = I_shm;
                rsbac_target_id.ipc.id.id_nr = inode->i_ino;
              }
            else
              {
                rsbac_target = T_FILE;
                rsbac_target_id.file.device = inode->i_dev;
                rsbac_target_id.file.inode  = inode->i_ino;
                rsbac_target_id.file.dentry_p = file->f_dentry;
              }
          }
        else
        if(S_ISBLK(inode->i_mode))
          {
            rsbac_target = T_DEV;
            rsbac_target_id.dev.type = D_block;
            rsbac_target_id.dev.id = inode->i_rdev;
          }
        else
        if(S_ISCHR(inode->i_mode))
          {
            rsbac_target = T_DEV;
            rsbac_target_id.dev.type = D_char;
            rsbac_target_id.dev.id = inode->i_rdev;
          }
        if(type == VERIFY_WRITE)
          rsbac_request = R_READ;
        else
          rsbac_request = R_WRITE;
        rsbac_attribute_value.dummy = 0;
        if (   (rsbac_target != T_NONE)
            && !rsbac_adf_request(rsbac_request,
                                  current->pid,
                                  rsbac_target,
                                  rsbac_target_id,
                                  A_none,
                                  rsbac_attribute_value)
           )
          {
            ret = -EPERM;
            goto out;
          }
        #endif /* CONFIG_RSBAC_RW */

	/* VERIFY_WRITE actually means a read, as we write to user space */
	fn = file->f_op->read;
	if (type == VERIFY_READ)
		fn = (io_fn_t) file->f_op->write;		

	ret = 0;
	vector = iov;
	while (count > 0) {
		void * base;
		size_t len;
		ssize_t nr;

		base = vector->iov_base;
		len = vector->iov_len;
		vector++;
		count--;

		nr = fn(file, base, len, &file->f_pos);

		if (nr < 0) {
			if (!ret) ret = nr;
			break;
		}
		ret += nr;
		if (nr != len)
			break;
	}

        /* RSBAC: notify adf of read or written file etc. */
        #ifdef CONFIG_RSBAC_RW
        if (   (ret > 0)
            && (rsbac_target != T_NONE)
           )
          {
            rsbac_new_target_id.dummy = 0;
            if (rsbac_adf_set_attr(rsbac_request,
                                   current->pid,
                                   rsbac_target,
                                   rsbac_target_id,
                                   T_NONE,
                                   rsbac_new_target_id,
                                   A_none,
                                   rsbac_attribute_value))
              {
                printk(KERN_WARNING
                       "do_readv_writev() [sys_vread(), sys_vwrite()]: rsbac_adf_set_attr() returned error\n");
              }
          }
        #endif

out:
	if (iov != iovstack)
		kfree(iov);
out_nofree:
	return ret;
}

asmlinkage ssize_t sys_readv(unsigned long fd, const struct iovec * vector,
			     unsigned long count)
{
	struct file * file;
	ssize_t ret;

	lock_kernel();

	ret = -EBADF;
	file = fget(fd);
	if (!file)
		goto bad_file;
	if (file->f_op && file->f_op->read && (file->f_mode & FMODE_READ))
		ret = do_readv_writev(VERIFY_WRITE, file, vector, count);
	fput(file);

bad_file:
	unlock_kernel();
	return ret;
}

asmlinkage ssize_t sys_writev(unsigned long fd, const struct iovec * vector,
			      unsigned long count)
{
	struct file * file;
	ssize_t ret;

	lock_kernel();

	ret = -EBADF;
	file = fget(fd);
	if (!file)
		goto bad_file;
	if (file->f_op && file->f_op->write && (file->f_mode & FMODE_WRITE)) {
		fs_down(&file->f_dentry->d_inode->i_sem);
		ret = do_readv_writev(VERIFY_READ, file, vector, count);
		fs_up(&file->f_dentry->d_inode->i_sem);
	}
	fput(file);

bad_file:
	unlock_kernel();
	return ret;
}

/* From the Single Unix Spec: pread & pwrite act like lseek to pos + op +
   lseek back to original location.  They fail just like lseek does on
   non-seekable files.  */

asmlinkage ssize_t sys_pread(unsigned int fd, char * buf,
			     size_t count, loff_t pos)
{
	ssize_t ret;
	struct file * file;
	ssize_t (*read)(struct file *, char *, size_t, loff_t *);
	struct inode * inode;

        /* RSBAC */
        #ifdef CONFIG_RSBAC_RW
        enum  rsbac_target_t          rsbac_target = T_NONE;
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_target_id_t       rsbac_new_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

	lock_kernel();

	ret = -EBADF;
	file = fget(fd);
	if (!file)
		goto bad_file;

	inode = file->f_dentry->d_inode;

	if (!(file->f_mode & FMODE_READ))
		goto out;

	ret = locks_verify_area(FLOCK_VERIFY_READ, inode, file, pos, count);
	if (ret)
		goto out;
	ret = -EINVAL;
	if (!file->f_op || !(read = file->f_op->read))
		goto out;
	if (pos < 0)
		goto out;

        /* RSBAC */
        #ifdef CONFIG_RSBAC_RW
#ifdef CONFIG_RSBAC_DEBUG
        if (rsbac_debug_aef)
          printk(KERN_DEBUG "sys_pread(): calling ADF\n");
#endif
        /* get target type and id clear */
        if(S_ISREG(inode->i_mode))
          {
            /* SHM special files */
            if (   inode->i_sb
                && (inode->i_sb->s_magic == SHM_FS_MAGIC)
               )
              {
                rsbac_target = T_IPC;
                rsbac_target_id.ipc.type = I_shm;
                rsbac_target_id.ipc.id.id_nr = inode->i_ino;
              }
            else
              {
                rsbac_target = T_FILE;
                rsbac_target_id.file.device = inode->i_dev;
                rsbac_target_id.file.inode  = inode->i_ino;
                rsbac_target_id.file.dentry_p = file->f_dentry;
              }
          }
        else
        if(S_ISBLK(inode->i_mode))
          {
            rsbac_target = T_DEV;
            rsbac_target_id.dev.type = D_block;
            rsbac_target_id.dev.id = inode->i_rdev;
          }
        else
        if(S_ISCHR(inode->i_mode))
          {
            rsbac_target = T_DEV;
            rsbac_target_id.dev.type = D_char;
            rsbac_target_id.dev.id = inode->i_rdev;
          }
        rsbac_attribute_value.dummy = 0;
        if (   (rsbac_target != T_NONE)
            && !rsbac_adf_request(R_READ,
                                  current->pid,
                                  rsbac_target,
                                  rsbac_target_id,
                                  A_none,
                                  rsbac_attribute_value)
           )
          {
            ret = -EPERM;
            goto out;
          }
        #endif /* CONFIG_RSBAC_RW */

	ret = read(file, buf, count, &pos);

        /* RSBAC: notify adf of read file etc. */
        #ifdef CONFIG_RSBAC_RW
        if (   (ret > 0)
            && (rsbac_target != T_NONE)
           )
          {
            rsbac_new_target_id.dummy = 0;
            if (rsbac_adf_set_attr(R_READ,
                                   current->pid,
                                   rsbac_target,
                                   rsbac_target_id,
                                   T_NONE,
                                   rsbac_new_target_id,
                                   A_none,
                                   rsbac_attribute_value))
              {
                printk(KERN_WARNING
                       "sys_pread(): rsbac_adf_set_attr() returned error\n");
              }
          }
        #endif

out:
	fput(file);
bad_file:
	unlock_kernel();
	return ret;
}

asmlinkage ssize_t sys_pwrite(unsigned int fd, const char * buf,
			      size_t count, loff_t pos)
{
	ssize_t ret;
	struct file * file;
	ssize_t (*write)(struct file *, const char *, size_t, loff_t *);
	struct inode * inode;

        /* RSBAC */
        #ifdef CONFIG_RSBAC_RW
        enum  rsbac_target_t          rsbac_target = T_NONE;
        union rsbac_target_id_t       rsbac_target_id;
        union rsbac_target_id_t       rsbac_new_target_id;
        union rsbac_attribute_value_t rsbac_attribute_value;
        #endif

	lock_kernel();

	ret = -EBADF;
	file = fget(fd);
	if (!file)
		goto bad_file;
	if (!(file->f_mode & FMODE_WRITE))
		goto out;

	inode = file->f_dentry->d_inode;

	ret = locks_verify_area(FLOCK_VERIFY_WRITE, inode, file, pos, count);
	if (ret)
		goto out;
	ret = -EINVAL;
	if (!file->f_op || !(write = file->f_op->write))
		goto out;
	if (pos < 0)
		goto out;

        /* RSBAC */
        #ifdef CONFIG_RSBAC_RW
#ifdef CONFIG_RSBAC_DEBUG
        if (rsbac_debug_aef)
          printk(KERN_DEBUG "sys_pwrite(): calling ADF\n");
#endif
        /* get target type and id clear */
        if(S_ISREG(inode->i_mode))
          {
            /* SHM special files */
            if (   inode->i_sb
                && (inode->i_sb->s_magic == SHM_FS_MAGIC)
               )
              {
                rsbac_target = T_IPC;
                rsbac_target_id.ipc.type = I_shm;
                rsbac_target_id.ipc.id.id_nr = inode->i_ino;
              }
            else
              {
                rsbac_target = T_FILE;
                rsbac_target_id.file.device = inode->i_dev;
                rsbac_target_id.file.inode  = inode->i_ino;
                rsbac_target_id.file.dentry_p = file->f_dentry;
              }
          }
        else
        if(S_ISBLK(inode->i_mode))
          {
            rsbac_target = T_DEV;
            rsbac_target_id.dev.type = D_block;
            rsbac_target_id.dev.id = inode->i_rdev;
          }
        else
        if(S_ISCHR(inode->i_mode))
          {
            rsbac_target = T_DEV;
            rsbac_target_id.dev.type = D_char;
            rsbac_target_id.dev.id = inode->i_rdev;
          }
        rsbac_attribute_value.dummy = 0;
        if (   (rsbac_target != T_NONE)
            && !rsbac_adf_request(R_WRITE,
                                  current->pid,
                                  rsbac_target,
                                  rsbac_target_id,
                                  A_none,
                                  rsbac_attribute_value)
           )
          {
            ret = -EPERM;
            goto out;
          }
        #endif /* CONFIG_RSBAC_RW */

	fs_down(&inode->i_sem);
	ret = write(file, buf, count, &pos);
	fs_up(&inode->i_sem);

        /* RSBAC: notify adf of written file etc. */
        #ifdef CONFIG_RSBAC_RW
        if (   (ret > 0)
            && (rsbac_target != T_NONE)
           )
          {
            rsbac_new_target_id.dummy = 0;
            if (rsbac_adf_set_attr(R_WRITE,
                                   current->pid,
                                   rsbac_target,
                                   rsbac_target_id,
                                   T_NONE,
                                   rsbac_new_target_id,
                                   A_none,
                                   rsbac_attribute_value))
              {
                printk(KERN_WARNING
                       "sys_pwrite(): rsbac_adf_set_attr() returned error\n");
              }
          }
        #endif

out:
	fput(file);
bad_file:
	unlock_kernel();
	return ret;
}
