/*
 * INET		An implementation of the TCP/IP protocol suite for the LINUX
 *		operating system.  INET is implemented using the  BSD Socket
 *		interface as the means of communication with the user level.
 *
 *		Generic frame diversion
 *
 * Version:	@(#)eth.c	0.41	09/09/2000
 *
 * Authors:	
 * 		Benoit LOCHER:	initial integration within the kernel with support for ethernet
 * 		Dave Miller:	improvement on the code (correctness, performance and source files)
 *
 */
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/errno.h>
#include <linux/config.h>
#include <linux/init.h>
#include <net/dst.h>
#include <net/arp.h>
#include <net/sock.h>
#include <net/ipv6.h>
#include <net/ip.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/checksum.h>
#include <linux/divert.h>
#include <linux/sockios.h>


const char sysctl_divert_version[32]="0.5";	/* Current version */


__initfunc(void dv_init(void))
{
	printk(KERN_INFO "NET4: Frame Diverter %s\n", sysctl_divert_version);
}


/*
 * Allocate a divert_blk for a device. This must be an ethernet nic.
 */
 
int alloc_divert_blk(struct device *dev)
{
	int alloc_size=(sizeof(struct divert_blk)+3)&~3;
	
	if (!strncmp(dev->name, "eth", 3))
	{
		printk(KERN_DEBUG "divert: allocating divert_blk for %s\n",
			dev->name);
		dev->divert=(struct divert_blk *)kmalloc(alloc_size, GFP_KERNEL);
		if (dev->divert==NULL)
		{
			printk(KERN_DEBUG "divert: unable to allocate divert_blk for %s\n",
				dev->name);
			return -EFAULT;
		}
		else
		{
			dev->divert->rwlock= RW_LOCK_UNLOCKED;
			dev->divert->divert=0;
			dev->divert->rule=NULL;
		}
	}
	else
	{
		printk(KERN_DEBUG "divert: not allocating divert_blk for non-ethernet device %s\n",
			dev->name);
		dev->divert=NULL;
	}
	return 0;
}



/*
 * Free a divert_blk allocated by the above function, if it was
 * allocated on that device.
 *
 */
void free_divert_blk(struct device *dev)
{
	if (dev->divert)
	{
		kfree(dev->divert);
		dev->divert=NULL;
		printk(KERN_DEBUG "divert: freeing divert_blk for %s\n",
				dev->name);
	}
	else
	{
		printk(KERN_DEBUG "divert: no divert_blk to free, %s not ethernet\n",
				dev->name);
	}
	return;
}                                                                                                                                     

struct divert_rule *divert_alloc_rule(void)
{
	struct divert_rule	*rule;
	int			alloc_size=(sizeof(struct divert_rule)+3)&~3;

	rule=(struct divert_rule *)kmalloc(alloc_size, GFP_KERNEL);
	if (rule==NULL)
		return NULL;
	memset(rule, 0, alloc_size);
	rule->prev=NULL;
	rule->next=NULL;
	return rule;
}


void divert_free_rule(struct divert_rule *rule)
{
	kfree(rule);
}


int divert_append_rule(struct device *dev, struct divert_rule *new_rule)
{
	struct divert_rule	*rule;
	unsigned long 		flags;

	if (dev->divert==NULL)
		return -EFAULT;
	
	/* write-lock the rules */
	write_lock_irqsave(&dev->divert->rwlock, flags);
	
	/* special case, 1st rule appended */
	new_rule->next=NULL;
	if (dev->divert->rule==NULL)
	{
		new_rule->prev=NULL;
		dev->divert->rule=new_rule;
	}
	else
	{
		rule=dev->divert->rule;
	
		while(rule->next!=NULL)
			rule=rule->next;
		rule->next=new_rule;
		new_rule->prev=rule;
		
	}
	
	/* release the write-lock on the rules */
	write_unlock_irqrestore(&dev->divert->rwlock, flags);
	
	return 0;
}


int divert_insert_rule(struct device *dev, struct divert_rule *new_rule, int position)
{
	struct divert_rule      *rule; 
	unsigned long			flags;
	
	if (dev->divert==NULL)
		return -EFAULT;

	/* write-lock the rules */
	write_lock_irqsave(&dev->divert->rwlock, flags);
	
	if (position==1 && dev->divert->rule==NULL)
	{
		dev->divert->rule=new_rule;
		new_rule->prev=NULL;
		new_rule->next=NULL;
		
	}
	else
	{
		rule=divert_get_rule(dev, position);
		if (rule==NULL)
		{
			/* release the write-lock */
			write_unlock_irqrestore(&dev->divert->rwlock, flags);
			
			return -EINVAL;
		}
		new_rule->next=rule;
		new_rule->prev=rule->prev;
		rule->prev=new_rule;
	}
	
	/* release the write-lock */
	write_unlock_irqrestore(&dev->divert->rwlock, flags);
	
	return 0;
}


int divert_delete_rule(struct device *dev, int position)
{
	struct divert_rule	*rule;
	unsigned long		flags;

	/* write-lock the rules */
	write_lock_irqsave(&dev->divert->rwlock, flags);
	
	rule=divert_get_rule(dev, position);
	if (rule==NULL)
	{
		write_unlock_irqrestore(&dev->divert->rwlock, flags);
		return -EINVAL;
	}

	/* First rule in the chain */
	if (rule->prev==NULL)
	{
		dev->divert->rule=rule->next;
		dev->divert->rule->prev=NULL;
	}
	/* Last rule in the chain */
	else if (rule->next==NULL)
	{
		rule->prev->next=NULL;
	}
	else
	{
		rule->prev->next=rule->next;
		rule->next->prev=rule->prev;
	}
	divert_free_rule(rule);
	
	/* release the write-lock on the rules */
	write_unlock_irqrestore(&dev->divert->rwlock, flags);
	
	return 0;
}


int divert_replace_rule(struct device *dev, struct divert_rule *new_rule, int position)
{
	struct divert_rule	*rule;
	unsigned long		flags;

	/* write-lock the rules */
	write_lock_irqsave(&dev->divert->rwlock, flags);
	
	rule=divert_get_rule(dev, position);
	if (rule==NULL)
	{
		/* release the write-lock on the rules */
		write_unlock_irqrestore(&dev->divert->rwlock, flags);
		
		return -EINVAL;
	}
	rule->next->prev=new_rule;
	rule->prev->next=new_rule;
	divert_free_rule(rule);

	/* release the write-lock */
	write_unlock_irqrestore(&dev->divert->rwlock, flags);
	
	return 0;
}
	

int divert_flush_rules(struct device *dev)
{
	struct divert_rule	*rule,
						*next_rule;
	unsigned long		flags;

	/* No diversion possible on this interface ? */
	if (!dev->divert)
		return -EINVAL;

	/* there are no rule set on this interface */
	if (!dev->divert->rule)
		return -EINVAL;

	/* Before flushing the rules, diversion on this nic must be off */
	if (dev->divert->divert)
		return -EINVAL;

	/* write-lock the rules */
	write_lock_irqsave(&dev->divert->rwlock, flags);
	
	rule=dev->divert->rule;
	dev->divert->rule=NULL;
	while(rule!=NULL)
	{
		next_rule=rule->next;
		divert_free_rule(rule);
		rule=next_rule;
	}

	/* release the write-lock */
	write_unlock_irqrestore(&dev->divert->rwlock, flags);
	
	return 0;
}


/*
 *
 * Must be called with divert->rwlock locked
 *
 */
struct divert_rule *divert_get_rule(struct device *dev, int rule_num)
{
	struct divert_rule	*rule;
	int		num=1;

	if (dev==NULL)
		return NULL;

	if (dev->divert==NULL)
		return NULL;
	
	rule=dev->divert->rule;
	while(rule!=NULL && rule_num!=num)
	{
		num++;
		rule=rule->next;
	}
	return	rule;
}


/* Some basic sanity checks on the arguments passed to divert_ioctl() */
int check_args(struct divert_cf *div_cf, struct device **dev)
{
	char    devname[32];
	
	if (dev==NULL)
		return -EFAULT;

	/* Let's try to find the ifname */
	*dev=dev_get(div_cf->dev_name);

	/* dev should NOT be null */
	if (*dev==NULL)
		return -EINVAL;
	
	/* user issuing the ioctl must be a super one :) */
	if (!suser())
		return -EPERM;

	/* Device must have a divert_blk member NOT null */
	if ((*dev)->divert==NULL)
		return -EFAULT;

	return 0;
}


int fill_rule_from_arg(struct divert_cf *div_cf, struct divert_rule *rule)
{
	rule->proto_num=div_cf->proto_num;
	memcpy((void *)&rule->proto_blk, (void *)&div_cf->proto_blk, sizeof(divert_proto));
	rule->divert_locally=div_cf->divert_locally;
	memcpy(rule->dest_mac, div_cf->dest_mac, ETH_ALEN);
	return 0;
}


int fill_arg_from_rule(struct divert_rule *rule, struct divert_cf *div_cf)
{
	div_cf->proto_num=rule->proto_num;
	memcpy((void *)&div_cf->proto_blk, (void *)&rule->proto_blk, sizeof(divert_proto));
	div_cf->divert_locally=rule->divert_locally;
	memcpy(div_cf->dest_mac, rule->dest_mac, ETH_ALEN);
	return 0;
}


int divert_ioctl(unsigned int cmd, struct divert_cf *arg)
{
	struct divert_cf	div_cf;
	struct device		*dev;
	int					oper, ret;
	struct divert_rule	*rule;

	if (copy_from_user(&div_cf, arg, sizeof(struct divert_cf)))
		return -EFAULT;
	
	ret=check_args(&div_cf, &dev);
	if (ret)
		return ret;
	
	oper=div_cf.cmd;
	
	switch(cmd)
	{
		case SIOCGIFDIVERT:
			switch(oper)
			{
				case DIVERT_GET_RULE:
					rule=divert_get_rule(dev, div_cf.rule_num);
					if (rule==NULL)
						return -EINVAL;
					ret=fill_arg_from_rule(rule, &div_cf);
					if (ret)
						return ret;
					break;

				case DIVERT_GET_STATUS:
					div_cf.status=dev->divert->divert;
					break;

				default:
					return -EINVAL;
			}
			if (copy_to_user(arg, &div_cf ,sizeof(struct divert_cf)))
				return -EFAULT;
			break;

		case SIOCSIFDIVERT:
			switch(oper)
			{
				case DIVERT_SET_STATE:
					if (div_cf.status && !dev->divert->divert)
					{
						dev->divert->divert=!dev->divert->divert;
						return 0;
					}
					else
						return -EALREADY;
					break;

				case DIVERT_APPEND_RULE:
					rule=divert_alloc_rule();
					if (rule==NULL)
						return -EFAULT;
					ret=fill_rule_from_arg(&div_cf, rule);
					if (ret)
					{
						/* Cleanup allocated memory */
						divert_free_rule(rule);
						return ret;
					}
					ret=divert_append_rule(dev, rule);
					if (ret)
					{
						/* Cleanup allocated memory */
						divert_free_rule(rule);
						return ret;
					}
					break;

				case DIVERT_INSERT_RULE:
					rule=divert_alloc_rule();
					if (rule==NULL)
						return -EFAULT;
					ret=fill_rule_from_arg(&div_cf, rule);
					if (ret)
					{
						/* Cleanup allocated memory */
						divert_free_rule(rule);
						return ret;
					}
					ret=divert_insert_rule(dev, rule, div_cf.rule_num);
					if (ret)
					{
						/* Cleanup allocated memory */
						divert_free_rule(rule);
						return ret;
					}
					break;

				case DIVERT_REPLACE_RULE:
					rule=divert_alloc_rule();
					if (rule==NULL)
						return -EFAULT;
					ret=fill_rule_from_arg(&div_cf, rule);
					if (ret)
					{
						/* Cleanup allocated memory */
						divert_free_rule(rule);
						return ret;
					}
					ret=divert_replace_rule(dev, rule, div_cf.rule_num);
					if (ret)
					{
						/* Cleanup allocated memory */
						divert_free_rule(rule);
						return ret;
					}
					break;

				case DIVERT_DELETE_RULE:
					ret=divert_delete_rule(dev, div_cf.rule_num);
					if (ret)
						return ret;
					break;

				default:
					return -EINVAL;
			}
			break;

		default:
			return -EINVAL;
	}
	return 0;
}


/*
 * Check if packet should have its dest mac address set to the box itself
 * for diversion
 */

#define	ETH_DIVERT_FRAME(skb) \
	memcpy(skb->mac.ethernet, skb->dev->dev_addr, ETH_ALEN); \
	skb->pkt_type=PACKET_HOST
	

#define	CHECK_CIDR(ip, cidr) (((ip)&(cidr.ipmask))==(cidr.ipnum))
	
	
int process_ip_rule(struct sk_buff *skb, struct divert_rule *rule)
{
	struct iphdr	*iph;
	unsigned char	*skb_data_end=skb->data+skb->len;
	
	iph=(struct iphdr *)(skb->data);
	/* Check for possible (maliciously) malformed IP frame (thanks Dave) */
	if (((iph->ihl<<2)+(unsigned char*)(iph))>=skb_data_end)
	{
		printk(KERN_INFO "divert: malformed IP packet !\n");
		return 0;	/* don't eat the frame */
	}
	if (!CHECK_CIDR(iph->saddr, rule->proto_blk.ip.scidr) ||
	    !CHECK_CIDR(iph->daddr, rule->proto_blk.ip.dcidr))
		return 0;
	if (rule->divert_locally)
			skb->pkt_type=PACKET_HOST;
	else
			memcpy(skb->mac.ethernet, rule->dest_mac, ETH_ALEN);
	return 1;	/* packet has been diverted */
}


int process_ip_tcp_rule(struct sk_buff *skb, struct divert_rule *rule)
{
	struct iphdr    *iph;
	struct tcphdr	*tcph;
	unsigned char   *skb_data_end=skb->data+skb->len;
	unsigned short  sport, dport;
	
	iph=(struct iphdr *)(skb->data);
	/* Check for possible (maliciously) malformed IP frame (thanks Dave) */
	if (((iph->ihl<<2)+(unsigned char*)(iph))>=skb_data_end)
	{
		printk(KERN_INFO "divert: malformed IP packet !\n");
		return 0;       /* don't eat the frame */
	}
	tcph=(struct tcphdr *)(((unsigned char *)iph) + (iph->ihl<<2));
	/* Check for possible (maliciously) malformed IP frame (thanx Dave) */
	if (((unsigned char *)(tcph+1))>=skb_data_end)
	{
		printk(KERN_INFO "divert: malformed TCP packet !\n");
		return 0;
	}
	sport=ntohs(tcph->source);
	dport=ntohs(tcph->dest);
	if (!CHECK_CIDR(iph->saddr, rule->proto_blk.tcp.scidr) ||
	    !CHECK_CIDR(iph->daddr, rule->proto_blk.tcp.dcidr))
		return 0; /* ip addresses do not match */
	if (sport<rule->proto_blk.tcp.sport_start ||
	    sport>rule->proto_blk.tcp.sport_end   ||
	    sport<rule->proto_blk.tcp.dport_start ||
	    sport>rule->proto_blk.tcp.sport_end)
		return 0; /* ports do not match */
	if (rule->divert_locally)
		skb->pkt_type=PACKET_HOST;
	else
		memcpy(skb->mac.ethernet, rule->dest_mac, ETH_ALEN);
	return 1;	/* packet has been diverted */
}


int process_ip_udp_rule(struct sk_buff *skb, struct divert_rule *rule)
{
	struct iphdr    *iph;
	struct udphdr	*udph;
	unsigned char   *skb_data_end=skb->data+skb->len;
	unsigned short  sport, dport;
	
	iph=(struct iphdr *)(skb->data);
	/* Check for possible (maliciously) malformed IP frame (thanks Dave) */
	if (((iph->ihl<<2)+(unsigned char*)(iph))>=skb_data_end)
	{
		printk(KERN_INFO "divert: malformed IP packet !\n");
		return 0;       /* don't eat the frame */
	}
	udph=(struct udphdr *)(((unsigned char *)iph) + (iph->ihl<<2));
	/* Check for possible (maliciously) malformed IP packet (thanks Dave) */
	if (((unsigned char *)(udph+1))>=skb_data_end)
	{
		printk(KERN_INFO "divert: malformed UDP packet !\n");
		return 0;	/* don't eat the frame */
	}
	sport=ntohs(udph->source);
	dport=ntohs(udph->dest);
	if (!CHECK_CIDR(iph->saddr, rule->proto_blk.udp.scidr) ||
	    !CHECK_CIDR(iph->daddr, rule->proto_blk.udp.dcidr))
		return 0; /* ip addresses do not match */
	if (sport<rule->proto_blk.udp.sport_start ||
	    sport>rule->proto_blk.udp.sport_end   ||
	    sport<rule->proto_blk.udp.dport_start ||
	    sport>rule->proto_blk.udp.sport_end)
		return 0; /* ports do not match */
	if (rule->divert_locally)
		skb->pkt_type=PACKET_HOST;
	else
		memcpy(skb->mac.ethernet, rule->dest_mac, ETH_ALEN);
	return 1;       /* packet has been diverted */
}

	
int process_ip_icmp_rule(struct sk_buff *skb, struct divert_rule *rule)
{
	struct iphdr    *iph;
	struct icmphdr	*icmph;
	unsigned char   *skb_data_end=skb->data+skb->len;
	unsigned short  type;

	iph=(struct iphdr *)(skb->data);
	/* Check for possible (maliciously) malformed IP frame (thanks Dave) */
	if (((iph->ihl<<2)+(unsigned char*)(iph))>=skb_data_end)
	{
		printk(KERN_INFO "divert: malformed IP packet !\n");
		return 0; /* don't eat the packet */
	}
	icmph=(struct icmphdr *)(((unsigned char *)iph) + (iph->ihl<<2));
	/* Check for possible (maliciously) malformed IP packet (thanks Dave) */
	if (((unsigned char *)(icmph+1))>=skb_data_end)
	{
		printk(KERN_INFO "divert: malformed ICMP packet !\n");
		return 0; /* don't eat the frame */
	}
	if (!CHECK_CIDR(iph->saddr, rule->proto_blk.icmp.scidr) ||
	    !CHECK_CIDR(iph->daddr, rule->proto_blk.icmp.dcidr))
		return 0; /* ip addresses do not match */
	type=icmph->type;
	if (type!=rule->proto_blk.icmp.type)
		return 0; /* icmp types do not match */
	if (rule->divert_locally)
		skb->pkt_type=PACKET_HOST;
	else
		memcpy(skb->mac.ethernet, rule->dest_mac, ETH_ALEN);
	return 1;
}


void divert_frame(struct sk_buff *skb)
{
	struct ethhdr			*eth=skb->mac.ethernet;
	struct divert_rule		*rule=skb->dev->divert->rule;
	unsigned long			flags;

	/* read-lock the rule list for this interface */
	read_lock_irqsave(&skb->dev->divert->rwlock, flags);

	while(rule)
	{
		if (rule->proto_num==eth->h_proto)
		{
			/*
			 * if rule_process() returns !=0 then stop further 
			 * diversion on this packet 
			 */
			if (rule->process_rule(skb, rule))
				break;
		}
		rule=rule->next;

	}

	/* release the read lock on this interface */
	write_unlock_irqrestore(&skb->dev->divert->rwlock, flags);
}
